/*
 * linux/drivers/char/amd_flash.c  Version 0.1
 *
 **********************************************************
 * NOTE:
 *
 *  This driver has only been verified with an Am29LV017B 4x8bit configuration.
 *
 **********************************************************
 * 
 * Copyright (C) 1999 Alexander Larsson (alla@lysator.liu.se or alex@signum.se)
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/malloc.h> 
#include <linux/init.h> 
#include <linux/time.h>
#include <linux/flash.h>

#include <asm/uaccess.h>

//#define DEBUGFLASH

#ifdef DEBUGFLASH
#define DEBUG(args...) printk(args)
#else
#define DEBUG(args...)
#endif

static int amd_read(struct flash_device_info *, char *, size_t, off_t, int );
static int amd_write(struct flash_device_info *, const char *, size_t, off_t, int);
static int amd_erase_sector(struct flash_device_info *, off_t, size_t);
static int amd_protect_sector(struct flash_device_info *, off_t, size_t, int);
static int amd_get_info(struct flash_device_info *, flash_info_t *);

struct amd_flash {
  volatile char *ptr;
  u_long size;
  short bit_width; /* 0==8bit, 1==16bit, 2==32bit */
  short chip_cfg; /* 0==1 chip, 1== 2 chip, 2== 4 chip */
  char erase_suspend; /* 0=unsupported, 1=read only, 2=r/w */
  struct semaphore mutex;
  /* etc.. sector protect etc... */
};
#define individual_bit_width(flash) ((flash)->bit_width-(flash)->chip_cfg)

struct flash_device_ops amd_ops = {
  amd_read,
  amd_write,
  amd_erase_sector,
  amd_protect_sector,
  amd_get_info
};

void __init amd_flash_init(char *ptr, u_long size, int bit_width, int chip_cfg) 
{
  struct amd_flash *flash;
  struct flash_device_info *fdi;
  int i;
  u_long offset;
  u_long reg_info;
  int sector_size;
  
  flash = kmalloc(sizeof(struct amd_flash), GFP_KERNEL);
  if (!flash)
    return;

  fdi = kmalloc(sizeof(struct flash_device_info), GFP_KERNEL);
  if (!fdi) {
    kfree(flash);
    return;
  }

  fdi->num_regions = cfi_read_char(ptr, CFI_NUM_ERASE_BLOCKS_OFFSET,
                                     0, bit_width);
  DEBUG("Num regions: %d\n", fdi->num_regions);
  
  fdi->region = kmalloc(sizeof(flash_region_info_t)*fdi->num_regions, GFP_KERNEL);
  if (!fdi->region) {
    kfree(flash);
    kfree(fdi);
    return;
  }

  flash->ptr = ptr;
  flash->size = size;
  flash->bit_width = bit_width;
  flash->chip_cfg = chip_cfg;
  flash->mutex = MUTEX;
    
  fdi->ops = &amd_ops;
  fdi->next = NULL;
  fdi->handle = flash;
  fdi->dev = 0;
  
  offset = 0;
  for (i=0;i<fdi->num_regions;i++) {
    reg_info = cfi_read_int(ptr, CFI_ERASE_REGION_INFO_OFFSET,
                            0, bit_width, chip_cfg);
    fdi->region[i].offset = offset;
    sector_size = (reg_info >> 16) & 0xFFFF;
    if (sector_size==0)
      fdi->region[i].sector_size = 128<<chip_cfg;
    else 
      fdi->region[i].sector_size = (sector_size*256)<<chip_cfg;
    
    fdi->region[i].num_sectors = (reg_info & 0xFFFF) + 1;

    DEBUG("sector %i size: %d, num_sectors: %d, total size: %d\n", i,
           fdi->region[i].sector_size , fdi->region[i].num_sectors,
           fdi->region[i].sector_size * fdi->region[i].num_sectors);
    
    offset += fdi->region[i].sector_size * fdi->region[i].num_sectors;
  }

  register_flash(fdi);

  /* Return to read array mode: */

  ptr[0] = 0xff;
  ptr[1] = 0xff;
  ptr[2] = 0xff;
  ptr[3] = 0xff;
}

static u_int read_with_chip_width(struct amd_flash *flash, u_long offs)
{
  u_int data;
  volatile char *ptr = flash->ptr;
  
  switch(individual_bit_width(flash)) {
  case 0: /* 8bit */
    data = (ptr[offs]);
    break;
  case 1: /* 16bit */
    data = (*(u_short *)&ptr[offs]);
    break;
  case 2: /* 32bit */
    data = (*(u_int *)&ptr[offs]);
    break;
  }
  return data;
}

#define PROGRAM_WRITE 1
#define PROGRAM_ERASE 2

static int program_in_progress(struct amd_flash *flash)
{
  /* This function doesn't work reliably if erase_suspend is used. */
  volatile char *ptr;
  int i;
  int num_chip;
  u_int data1, data2, data;
  
  num_chip = 1<<flash->chip_cfg;

  ptr = flash->ptr;

  for (i=0;i<num_chip;i++) {
    data1 = read_with_chip_width(flash, 0) & 0x44; /* mask out DQ6 and DQ2 */
    data2 = read_with_chip_width(flash, 0) & 0x44 ;/* mask out DQ6 and DQ2 */
    data = data1 ^ data2; /* xor */
    if (data&0x40) {
      /* DQ6 toggled, some program is running (might be in erase suspend) */
      DEBUG("dq6 on chip %d toggled, embedded program running.\n", i);
      if (data&0x04) {
        DEBUG("dq2 on chip %d toggled, erase runnung.\n", i);
        return PROGRAM_ERASE;
      } else {
        DEBUG("dq2 on chip %d not toggled, write runnung.\n", i);
        return PROGRAM_WRITE;
      }
    }
  }
  return 0; /* No program in progress. */
}

static void do_erase_sector(volatile char *ptr, off_t offset)
{
  u_long j;
  u_int data;

  j = jiffies;
  
  *(volatile int *)&ptr[offset] = 0xaaaaaaaa;
  *(volatile int *)&ptr[offset] = 0x55555555;
  *(volatile int *)&ptr[offset] = 0x80808080;
  *(volatile int *)&ptr[offset] = 0xaaaaaaaa;
  *(volatile int *)&ptr[offset] = 0x55555555;
  *(volatile int *)&ptr[offset] = 0x30303030;

  current->state = TASK_INTERRUPTIBLE;
  schedule_timeout(10*100/HZ); /* Time out 0.1 sec. */
  while ((data = *(volatile int *)&ptr[offset])!=0xffffffff) {
    /*    DEBUG("data= %x, delaying more...\n", data);*/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(10*100/HZ); /* Time out 0.1 sec. */
  }

  j = jiffies - j;
  DEBUG("Erasing %ld took %ld jiffies\n", offset, j);
}

static int amd_read(struct flash_device_info *fdi,
                    char *buf, size_t count, off_t offset, int userspace_ptr)
{
  struct amd_flash *flash = fdi->handle;
  int ret;

  ret = 0;

  /*  DEBUG("amd_read buf=%x, count=%d, offset=%x, us_ptr=%d\n",
      (int)buf, (int)count, (int)offset, (int) userspace_ptr);*/
  
  down(&flash->mutex);
  
  if (program_in_progress(flash)) {
    printk("Error, embedded program in progress during flash read\n");
    ret = -EIO;
    goto error;
  }

  if (userspace_ptr) {
    if (copy_to_user(buf, (char *)flash->ptr+offset, count) > 0) {
      ret = -EFAULT;
    }
  } else {
    memcpy(buf, (char *)flash->ptr+offset, count);
  }
  
 error:
 
  up(&flash->mutex);

  return ret;
}

static void unlock_bypass(volatile char *ptr, off_t offset)
{
  *(volatile int *)&ptr[offset] = 0xaaaaaaaa;
  *(volatile int *)&ptr[offset] = 0x55555555;
  *(volatile int *)&ptr[offset] = 0x20202020;
}

static void unlock_bypass_reset(volatile char *ptr, off_t offset)
{
  *(volatile int *)&ptr[offset] = 0x90909090;
  *(volatile int *)&ptr[offset] = 0x00000000;
}

static void do_write_int(volatile char *ptr, off_t offset, u_int data )
{
  int wait;

  if ( ((*(volatile u_int *)&ptr[offset])&data) != data) {
    printk("Flash write error, not fully erased at %lx\n", offset);
    data = (*(volatile u_int *)&ptr[offset])&data;
  }
  /* Program word: */
  *(volatile u_int *)&ptr[offset] = 0xa0a0a0a0;
  *(volatile u_int *)&ptr[offset] = data;

  wait=0;
  while ((*(volatile u_int *)&ptr[offset]) != data) {
    wait++;
    if (wait==1000) {
      DEBUG("do_write_int, wait==1000!\n");
      DEBUG("offet = %lx, data = %x, flash: %x\n",
              offset, data, (*(volatile u_int *)&ptr[offset])); 
    }
    /* Busy wait */
  }

}

static void do_write_char(volatile u_char *ptr, off_t offset, u_char data )
{
  int wait;
  if ( ((ptr[offset])&data) != data) {
    printk("Flash write error, not fully erased at %lx\n", offset);
    data = (ptr[offset])&data;
  }
  /* Program byte: */
  ptr[offset] = 0xa0;
  ptr[offset] = data;
  
  wait=0;
  while (ptr[offset] != data) {
    wait++;
    if (wait==1000) {
      DEBUG("do_write_char, wait==1000!\n");
      DEBUG("offet = %lx, data = %x, flash: %x\n",
              offset, (u_int) data, (u_int) ptr[offset]); 
    }
    /* Busy wait */
  }
}


static int amd_write(struct flash_device_info *fdi,
                     const char *buf, size_t count, off_t offset, int userspace_ptr)
{
  struct amd_flash *flash = fdi->handle;
  struct timeval time1, time2;
  volatile char *ptr;
  int ret;
  u_long end_offset;
  u_int data;
  int loops;
  
  ret = 0;
  ptr = flash->ptr;

  /*DEBUG("amd_write buf=%x, count=%d, offset=%x, us_ptr=%d\n",
    (int)buf, (int)count, (int)offset, (int) userspace_ptr);*/
  
  down(&flash->mutex);
  
  if (program_in_progress(flash)) {
    printk("Error, embedded program in progress during flash read\n");
    ret = -EIO;
    goto error;
  }
  
  end_offset = offset + count;
  
  if (userspace_ptr && (!access_ok(VERIFY_READ, buf, count)) ) {
    ret = -EFAULT;
    goto error;
  }

  unlock_bypass(ptr, offset);

  do_gettimeofday(&time1);
  
  loops = 0;
  while (offset<end_offset) {
    if ((offset%4 == 0) && (end_offset-offset >= 4)) {
      if (userspace_ptr) {
        if (__get_user(data, (int *)buf)) {
          ret = -EFAULT;
          goto unlock_bypass_reset_error;
        }
      } else {
        data = *(int *)buf;
      }

      do_write_int(ptr, offset, data);
      buf += 4;
      offset += 4;
    } else {
      if (userspace_ptr) {
        if (__get_user(data, (char *)buf)) {
          ret = -EFAULT;
          goto unlock_bypass_reset_error;
        }
      } else {
        data = *(char *)buf;
      }
      do_write_char(ptr, offset, data);
      buf += 1;
      offset += 1;
    }

    /* every 32 writes we check if >1msec has passed. then we schedule */
    if ((++loops & 0x1f) == 0) {
      do_gettimeofday(&time2);
      time2.tv_sec -= time1.tv_sec; 
      time2.tv_usec -= time1.tv_usec;

      if ((time2.tv_sec>0) || (time2.tv_usec>1000)) {
        /* May wait to short if tv_sec wraps. Who cares... */
        current->state = TASK_RUNNING;
        schedule();
        if (userspace_ptr && (!access_ok(VERIFY_READ, buf, count))) {
          ret = -EFAULT;
          goto unlock_bypass_reset_error;
        }
        do_gettimeofday(&time1);
      }
    }
  }
  
 unlock_bypass_reset_error:
  unlock_bypass_reset(ptr, offset);
 error:
  up(&flash->mutex);

  return ret;
}

static int amd_erase_sector(struct flash_device_info *fdi,
                            off_t offset, size_t count)
{
  struct amd_flash *flash = fdi->handle;
  flash_region_info_t *region = NULL;
  u_long region_size;
  u_long start, end;
  int ret;
  int i;

  ret = 0;

  /*DEBUG("amd_erase count=%d, offset=%x\n",
    (int)count, (int)offset);*/
  
  down(&flash->mutex);

  if (!flash_on_sector_boundaries(fdi, offset, count)) {
    DEBUG("Trying to erase, but not at sector boundaries. offs=%x, count=%x\n",
            (int)offset, (int)count);
    ret = -EINVAL;
    goto error;
  }

  if (program_in_progress(flash)) {
    printk("Error, embedded program in progress during flash erase\n");
    ret = -EIO;
    goto error;
  }

  /* erase all sectors in mem region */
  /* Now guaranteed to be on sector boundaries */
  for (i = 0; i<fdi->num_regions; i++) {
    region = &fdi->region[i];
    region_size = region->sector_size*region->num_sectors;
    if  ( (offset >= region->offset) &&
          (offset < region->offset+region_size) ) {
      start = offset - region->offset;
      end = start + count;
      
      if (end > region_size) {
        offset = region->offset+region_size;
        count -= region_size - end;
        end = region_size;
      }

      while (start<end) {
        do_erase_sector(flash->ptr, start);
        start += region->sector_size;
      }
    }
  }

  
  
 error:
 
  up(&flash->mutex);

  return ret;
}

static int amd_protect_sector(struct flash_device_info *fdi,
                              off_t offset, size_t count, int protected)
{
  return -ENOSYS;
}

static int amd_get_info(struct flash_device_info *fdi,
                         flash_info_t *flash_info)
{
  struct amd_flash *flash = fdi->handle;
  flash_info_t t;
  int ret;

  t.size = flash->size;
  t.bitwidth = 8<<individual_bit_width(flash);
  t.chip_config = 1<<flash->chip_cfg;
  t.num_regions = fdi->num_regions;
  t.num_regions = fdi->num_regions;

  ret = copy_to_user(flash_info, &t, sizeof(flash_info_t));
  if (ret)
    return -EFAULT;
  
  return 0;
}
