/*
 * linux/drivers/char/flash.c  Version 0.1
 *
 * Copyright (C) 1999 Alexander Larsson (alla@lysator.liu.se or alex@signum.se)
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h> 
#include <linux/init.h> 
#include <linux/flash.h>

#include <asm/uaccess.h>
#include <asm/io.h>

//#define DEBUGFLASH

#ifdef DEBUGFLASH
#define DEBUG(args...) printk(args)
#else
#define DEBUG(args...)
#endif

flash_device_info_t *flash_devices = NULL;
static int next_minor = 0;

struct flash_chip_data {
  u_long addr;
  u_long size;
};

static struct flash_chip_data probe_data[] __initdata = {
  { 0x10000000, 8*1024*1024},
};
#define NUM_FLASH_CHIP (sizeof(probe_data)/sizeof(struct flash_chip_data))


#define MAKE_TWO(byte) (((u_short)byte)<<8 | byte)
#define MAKE_FOUR(byte) (((u_int)byte)<<24 | byte<<16 | byte<<8 | byte)


int register_flash(struct flash_device_info *fdi)
{
  fdi->dev = MKDEV(FLASH_MAJOR, next_minor++);
  fdi->next = flash_devices;
  flash_devices = fdi;

  printk("registered flash device /dev/flash%d\n", MINOR(fdi->dev));
  
  return 0;
}

int unregister_flash(struct flash_device_info *unreg)
{
  struct flash_device_info *fdi, *prev;

  prev = NULL;
  fdi = flash_devices;
  while (fdi != NULL && fdi != unreg) {
    prev = fdi;
    fdi = fdi->next;
  }
  
  if (fdi == NULL)
    return -1;
  
  if (prev)
    prev->next = fdi->next;
  else
    flash_devices = fdi->next;

  printk("unregistered flash device\n");

  return 0;
}

static loff_t flash_lseek(struct file * file, loff_t offset, int orig)
{
  switch (orig) {
  case 0:
    file->f_pos = offset;
    return file->f_pos;
  case 1:
    file->f_pos += offset;
    return file->f_pos;
  default:
    return -EINVAL;
  }
}

static u_long end_of_flash(struct flash_device_info *fdi)
{
  flash_region_info_t *last_region;

  last_region = &fdi->region[fdi->num_regions-1];

  return last_region->offset +
    last_region->sector_size*last_region->num_sectors;
}

/*
 * This funcion reads the flash memory. 
 */
static ssize_t flash_read(struct file * file, char * buf,
                          size_t count, loff_t *ppos)
{
  struct flash_device_info *fdi = file->private_data;
  unsigned long p = *ppos;
  unsigned long end_flash;
  int res;

  end_flash = end_of_flash(fdi);
  
  if (p >= end_flash)
    return 0;
  
  if (count > end_flash - p)
    count = end_flash - p;

  res = (*fdi->ops->read)(fdi, buf, count, p, 1);

  if (res)
    return res;
  
  *ppos += count;
  return count;
}

static ssize_t flash_write(struct file * file, const char * buf, 
                           size_t count, loff_t *ppos)
{
  struct flash_device_info *fdi = file->private_data;
  unsigned long p = *ppos;
  unsigned long end_flash;
  int res;
  
  end_flash = end_of_flash(fdi);
  
  if (p >= end_flash)
    return 0;
  
  if (count > end_flash - p)
    count = end_flash - p;


  if (flash_on_sector_boundaries(fdi, p, count)) {
    /* On a sector boundary => Erase the sectors */
    res = (*fdi->ops->erase_sector)(fdi, p, count);
    if (res)
      return res;
  }
  
  res = (*fdi->ops->write)(fdi, buf, count, p, 1);
  if (res)
    return res;
  
  *ppos += count;
  return count;
}

static int flash_ioctl(struct inode * inode, struct file *filp,
                       unsigned int cmd, unsigned long arg)
{
  struct flash_device_info *fdi = filp->private_data;
  erase_info_t erase_info;
  int ret;
  
  switch (cmd) {
  case FLASHGETINFO:
    ret = (*fdi->ops->get_info)(fdi, (flash_info_t *)arg);
    return ret;
    break;
  case FLASHGETREGIONINFO:
    ret = copy_to_user((char *)arg, fdi->region,
                       sizeof(flash_region_info_t)*fdi->num_regions);
    if (ret)
      return -EFAULT;
    return 0;
    break;
  case FLASHERASE:
    ret = copy_from_user(&erase_info, (erase_info_t *)arg, sizeof(erase_info_t));
    if (ret)
      return ret;
    ret = (*fdi->ops->erase_sector)(fdi, erase_info.offset, erase_info.size);
    return ret;
    break;
  default:
    return -EINVAL;
  }
  return -EINVAL;
}

static
struct flash_device_info *flash_find_device (kdev_t dev)
{
  struct flash_device_info *fdi;

  fdi = flash_devices;
  while (fdi != NULL && fdi->dev != dev)
    fdi = fdi->next;
  return fdi;
}


static int flash_open(struct inode * inode, struct file * filp)
{
  kdev_t dev = inode->i_rdev;
  struct flash_device_info *fdi = flash_find_device(dev);

  DEBUG("entering flash_open, minor==%d\n", MINOR(dev));
  if (fdi == NULL)
    return -ENODEV;

  filp->private_data = fdi;

  DEBUG("opened device /dev/flash%d\n", MINOR(dev));

  return 0;
}


static struct file_operations flash_fops = {
	flash_lseek,	/* lseek */
	flash_read,	/* read */
	flash_write,	/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	flash_ioctl,	/* ioctl */
	NULL,		/* mmap */
	flash_open,	/* open */
	NULL,		/* flush */
	NULL,		/* release */
	NULL		/* fsync */
};

int flash_on_sector_boundaries(struct flash_device_info *fdi,
                              off_t offset, size_t count)
{
  flash_region_info_t *region = NULL;
  int i;
  u_long start, end;
  u_long region_size;

  for (i = 0; i<fdi->num_regions; i++) {
    region = &fdi->region[i];
    region_size = region->sector_size*region->num_sectors;
    if  ( (offset >= region->offset) &&
          (offset < region->offset+region_size) ) {
      start = offset - region->offset;
      end = start + count;
      
      if (start % region->sector_size != 0)
        return 0;  /* Not on sector boundaries */
      
      if (end > region_size) {
        offset = region->offset+region_size;
        count -= region_size - end;
        /* This region ok, Continue loop */
      } else {
        if (end % region->sector_size != 0)
          return 0;  /* Not on sector boundaries */
        else
          return 1; /* Both start and end on sector boundaries. */
      }
    }
  }

  return 0; /* Not on sector boundaries */
}



/********** Init stuff: ********/

#ifdef CONFIG_AMD_FLASH
extern void amd_flash_init(char *ptr, u_long size, int bit_width, int chip_cfg) __init;
#endif

static char *vendor_name_table[] __initdata = {
  "Unknown",
  "None", 
  "Intel/Sharp extended",
  "AMD/Fujitsu standard",
  "Intel standard",
  "AMD/Fujitsu extended",
  "Mitsubishi standard",
  "Mitsubishi extended"
  /* Add new at end */
};

static void __init cfi_flash_init(void)
{
  int i;
  char *ptr;
  char qry[6];
  int total_buswidth=0;
  int chip_buswidth=0;
  int found_flash;
  int bit_width;
  int chip_cfg;
  u_short primary_vendor;
  char *vendor_name;
  u_long size;

  for (i=0;i<NUM_FLASH_CHIP;i++) {
    DEBUG("Probing for chip %d at %x\n", i, (int)probe_data[i].addr );
    ptr = ioremap_nocache(probe_data[i].addr, probe_data[i].size);

    found_flash = 0;

    /* Probe at 0x10 for 8bit device chip */
    ptr[CFI_QUERY_OFFSET] = CFI_QUERY_CMD;
    qry[0] = ptr[0x10]; /* Extra read cycle */
    qry[0] = ptr[0x10];
    qry[1] = ptr[0x11];
    qry[2] = ptr[0x12];
    if ((qry[0]=='Q') && (qry[1]=='R') && (qry[2]=='Y')) {
      found_flash = 1;
      total_buswidth=8;
      chip_buswidth=8;
    }
    
    /* Probe at 0x20 for 16bit device chip */
    if (!found_flash) {
      *(u_short *)&ptr[CFI_QUERY_OFFSET<<1] = MAKE_TWO(CFI_QUERY_CMD);
      qry[0] = ptr[0x20]; /* Extra read cycle */
      qry[0] = ptr[0x20];
      qry[1] = ptr[0x21];
      qry[2] = ptr[0x22];
      qry[3] = ptr[0x23];
      qry[4] = ptr[0x24];
      qry[5] = ptr[0x25];
      if ((qry[0]=='Q') && (qry[1]=='Q') && (qry[2]=='R') &&
          (qry[3]=='R') && (qry[4]=='Y') && (qry[5]=='Y')) {
        found_flash = 1;
        total_buswidth=16;
        chip_buswidth=8;
      }
    }
    if (!found_flash) {
      *(u_short *)&ptr[CFI_QUERY_OFFSET<<1] = (u_short)CFI_QUERY_CMD;
      qry[0] = ptr[0x20]; /* Extra read cycle */
      qry[0] = ptr[0x20];
      qry[1] = ptr[0x21];
      qry[2] = ptr[0x22];
      qry[3] = ptr[0x23];
      qry[4] = ptr[0x24];
      qry[5] = ptr[0x25];
      if ((qry[0]=='Q') && (qry[1]== 0 ) && (qry[2]=='R') &&
          (qry[3]== 0 ) && (qry[4]=='Y') && (qry[5]== 0 )) {
        found_flash = 1;
        total_buswidth=16;
        chip_buswidth=16;
      }
    }
    
   /* Probe at 0x40 for 32bit device chip */
    if (!found_flash) {
      *(u_int *)&ptr[CFI_QUERY_OFFSET<<2] = MAKE_FOUR(CFI_QUERY_CMD);
      qry[0] = ptr[0x40]; /* Extra read cycle */
      qry[0] = ptr[0x40];
      qry[1] = ptr[0x41];
      qry[2] = ptr[0x42];
      qry[3] = ptr[0x43];
      qry[4] = ptr[0x44];
      qry[5] = ptr[0x45];
      if ((qry[0]=='Q') && (qry[1]=='Q') && (qry[2]=='Q') &&
          (qry[3]=='Q') && (qry[4]=='R') && (qry[5]=='R')) {
        found_flash = 1;
        total_buswidth=32;
        chip_buswidth=8;
      }
    }
    if (!found_flash) {
      *(u_int *)&ptr[CFI_QUERY_OFFSET<<2] = (u_int)CFI_QUERY_CMD<<16 | CFI_QUERY_CMD ;
      qry[0] = ptr[0x40]; /* Extra read cycle */
      qry[0] = ptr[0x40];
      qry[1] = ptr[0x41];
      qry[2] = ptr[0x42];
      qry[3] = ptr[0x43];
      qry[4] = ptr[0x44];
      qry[5] = ptr[0x45];
      if ((qry[0]=='Q') && (qry[1]== 0 ) && (qry[2]=='Q') &&
          (qry[3]== 0 ) && (qry[4]=='R') && (qry[5]== 0 )) {
        found_flash = 1;
        total_buswidth=32;
        chip_buswidth=16;
      }
    }
    if (!found_flash) {
      *(u_int *)&ptr[CFI_QUERY_OFFSET<<2] = (u_int)CFI_QUERY_CMD;
      qry[0] = ptr[0x40]; /* Extra read cycle */
      qry[0] = ptr[0x40];
      qry[1] = ptr[0x41];
      qry[2] = ptr[0x42];
      qry[3] = ptr[0x43];
      qry[4] = ptr[0x44];
      qry[5] = ptr[0x45];
      if ((qry[0]=='Q') && (qry[1]== 0 ) && (qry[2]== 0 ) &&
          (qry[3]== 0 ) && (qry[4]=='R') && (qry[5]== 0 )) {
        found_flash = 1;
        total_buswidth=32;
        chip_buswidth=32;
      }
    }

    if (!found_flash) {
      printk("Couldn't find flash device at %lx\n", probe_data[i].addr);
      iounmap(ptr);
      continue;
    }

    switch(total_buswidth) {
    case 32: bit_width = 2; break;
    case 16: bit_width = 1; break;
    case 8:  bit_width = 0; break;
    }
    switch(total_buswidth/chip_buswidth) {
    case 4: chip_cfg = 2; break;
    case 2: chip_cfg = 1; break;
    case 1: chip_cfg = 0; break;
    }

    printk("bit_width: %d, chip_cfg: %d\n", bit_width, chip_cfg);

    primary_vendor = cfi_read_short(ptr, CFI_PRIMARY_VENDOR_ID_OFFSET, 0,
                                    bit_width, chip_cfg);

    switch(primary_vendor) {
    case CFI_VENDOR_NONE:
      vendor_name = vendor_name_table[1];
    case CFI_VENDOR_INTEL_SHARP_EXTENDED:
      vendor_name = vendor_name_table[2];
      break;
    case CFI_VENDOR_AMD_FUJITSU_STANDARD:
      vendor_name = vendor_name_table[3];
      break;
    case CFI_VENDOR_INTEL_STANDARD:
      vendor_name = vendor_name_table[4];
      break;
    case CFI_VENDOR_AMD_FUJITSU_EXTENDED:
      vendor_name = vendor_name_table[5];
      break;
    case CFI_VENDOR_MITSUBISHI_STANDARD:
      vendor_name = vendor_name_table[6];
      break;
    case CFI_VENDOR_MITSUBISHI_EXTENDED:
      vendor_name = vendor_name_table[7];
      break;
    default:
      vendor_name = vendor_name_table[0];
      break;
    }

    size = cfi_read_char(ptr, CFI_DEVICE_SIZE_OFFSET, 0, bit_width);
    size = (1<<size) << chip_cfg;

    printk("Found %dx%dbit %dMbyte cfi flash device of type %s at %lx\n",
           total_buswidth/chip_buswidth, chip_buswidth,
           (int) (size/(1024*1024)),
           vendor_name, probe_data[i].addr);

    if (size!=probe_data[i].size) {
      printk("Detected size not same as probed. Resizeing.\n");
      iounmap(ptr);
      ptr = ioremap_nocache(probe_data[i].addr, size);
    }
    
#ifdef CONFIG_AMD_FLASH
    if (primary_vendor == CFI_VENDOR_AMD_FUJITSU_STANDARD)
      amd_flash_init(ptr, size, bit_width, chip_cfg);
#endif
    
  }
}

void __init flash_init(void)
{
  if (register_chrdev(FLASH_MAJOR,"flash",&flash_fops))
    printk("unable to get major %d for flash devs\n", FLASH_MAJOR);
  cfi_flash_init();

  /* probe for non-cfi flash memories are added here */
}








