
/* Linux driver for Disk-On-Chip 2000       */
/* (c) 1999 Machine Vision Holdings, Inc.   */
/* Author: David Woodhouse <dwmw2@mvhi.com> */
/* $Id: doc2000.c,v 1.1 1999/08/09 19:38:05 dwmw2 Exp $ */

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/sched.h>

#include <linux/types.h>
#include "mtd.h"

#ifdef __SMP__
#warning The DiskOnChip driver is not SMP safe. I need to add some locking -dw
#endif


static int doc_read (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf);
static ssize_t doc_write (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf);
static int doc_release(struct inode *inode, struct file *file);
static int doc_open(struct inode *inode, struct file *filp);
static int doc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int doc_read_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, u_char *buf);
//static int doc_read_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, u_char *buf);


#define DoC_M_CDSN_IO 0x800

#define DoC_ChipID 0x1000
#define DoC_DOCStatus 0x1001
#define DoC_DOCControl 0x1002
#define DoC_FloorSelect 0x1003
#define DoC_CSDNControl 0x1004
#define DoC_CSDNDeviceSelect 0x1005
#define DoC_ECCConf 0x1006

#define DoC_2k_Toggle 0x1007

#define DoC_CSDNSlowIO 0x100d
#define DoC_ECCSyndrome0 0x1010
#define DoC_ECCSyndrome1 0x1011
#define DoC_ECCSyndrome2 0x1012
#define DoC_ECCSyndrome3 0x1013
#define DoC_ECCSyndrome4 0x1014
#define DoC_ECCSyndrome5 0x1015
#define DoC_AliasResolution 0x101b
#define DoC_ConfigInput 0x101c
#define DoC_ReadPipeInit 0x101d
#define DoC_WritePipeTerm 0x101e
#define DoC_LastDataRead 0x101f
#define DoC_NOP 0x1020

#define DoC_2k_CSDN_IO 0x1800

#ifdef __i386__ /* i386 readb/writeb or the address with 0xc0000000 _every_ 
		   time. We don't need that, so we just reference it. */
#define ReadDOC(adr, reg)      *(((volatile unsigned char *)adr) + DoC_##reg)
#define WriteDOC(d, adr, reg)  (*(((volatile unsigned char *)adr) + DoC_##reg) = d)
#else
#define ReadDOC(adr, reg)      readb(((unsigned long)adr) + DoC_##reg)
#define WriteDOC(d, adr, reg)  writeb(d, ((unsigned long)adr) + DoC_##reg)
#endif

#define DOC_MODE_RESET 0
#define DOC_MODE_NORMAL 1
#define DOC_MODE_RESERVED1 2
#define DOC_MODE_RESERVED2 3

#define DOC_MODE_MDWREN 4
#define DOC_MODE_CLR_ERR 0x80

#define DOC_ChipID_Doc2k 0x20
#define DOC_ChipID_DocMil 0x30

#define CSDN_CTRL_FR_B 0x80
#define CSDN_CTRL_UNKNOWN 0x10
#define CSDN_CTRL_WP 8
#define CSDN_CTRL_ALE 4
#define CSDN_CTRL_CLE 2
#define CSDN_CTRL_CE 1

unsigned long doc_locations[] = { 0xc8000, 0xd0000, 0xd8000, 0 };

struct Nand {
  char floor, chip;
  unsigned long curadr;
  unsigned char curmode;
  /* Also some erase/write/pipeline info when we get that far */
};

#define MAX_FLOORS 4
#define MAX_CHIPS 4

struct DiskOnChip {
  unsigned long physadr;
  unsigned long virtadr;
  unsigned long totlen;
  char ChipID; /* Type of DiskOnChip */

  unsigned long mfr; /* Flash IDs - only one type of flash per device */
  unsigned long id;
  int chipshift;
  
  int curfloor;
  int curchip;

  int numchips;
  struct Nand *chips;
};

static int numdocs = 0;

int doccheck(unsigned long potential, unsigned long physadr);



int _DoC_WaitReady (unsigned long docptr)
{

  long c=-1;

  while (!(ReadDOC(docptr, CSDNControl) & CSDN_CTRL_FR_B) && --c)
    ;
  
  if (c == 0)
    printk(KERN_WARNING "_DoC_WaitReady timed out.");

  return (c==0);
}

static inline int DoC_WaitReady(unsigned long docptr) 
{

  /* This is inline, to optimise the case where it's ready instantly */

  if (!(ReadDOC(docptr, CSDNControl) & CSDN_CTRL_FR_B))
    return _DoC_WaitReady(docptr); /* Call the out-of-line routine and wait */


  return 0;
}

static inline int DoC_Command(unsigned long docptr, unsigned char command)
{
 	WriteDOC( /* Doc2000 driver used 0x1d but I don't know what the
		   0x10 bit does */
		 CSDN_CTRL_UNKNOWN | CSDN_CTRL_WP | CSDN_CTRL_CLE | CSDN_CTRL_CE,
		 docptr, CSDNControl);

	if (DoC_WaitReady(docptr))
	  return 1;

	WriteDOC(command, docptr, 2k_CSDN_IO);
	//		WriteDOC(0x00, docptr, IPLalias);

	WriteDOC( CSDN_CTRL_UNKNOWN | CSDN_CTRL_WP | CSDN_CTRL_CE, docptr, CSDNControl);

	return DoC_WaitReady(docptr);
}


static inline int DoC_Address (unsigned long docptr, unsigned long ofs)
{
 	WriteDOC( /* Doc2000 driver used 0x1d but I don't know what the
		   0x10 bit does */
		 CSDN_CTRL_UNKNOWN | CSDN_CTRL_WP | CSDN_CTRL_ALE | CSDN_CTRL_CE,
		 docptr, CSDNControl);

	if (DoC_WaitReady(docptr))
	    return 1;

	WriteDOC(ofs & 0xff, docptr, 2k_CSDN_IO);
	
	WriteDOC((ofs >> 9) & 0xff, docptr, 2k_CSDN_IO);

	WriteDOC((ofs >> 17) & 0xff, docptr, 2k_CSDN_IO);
	
	WriteDOC( CSDN_CTRL_UNKNOWN | CSDN_CTRL_WP | CSDN_CTRL_CE, docptr, CSDNControl);
	
	return DoC_WaitReady(docptr);
}

static inline int DoC_SelectChip(unsigned long docptr, int chip)
{
  WriteDOC( chip, docptr, CSDNDeviceSelect);
  return DoC_WaitReady(docptr);
}

static inline int DoC_SelectFloor(unsigned long docptr, int floor)
{
  WriteDOC( floor, docptr, FloorSelect);
  return DoC_WaitReady(docptr);
}
  
int DoC_IdentChip(struct DiskOnChip *doc, int floor, int chip)
{
  int mfr, id, chipshift=0;
  char *mfrname=NULL, *idname=NULL;
  DoC_SelectFloor(doc->virtadr, floor);
  DoC_SelectChip(doc->virtadr, chip);

  if (DoC_Command(doc->virtadr, 0xff)) {
    printk("DoC_Command (reset) for %d,%d returned true\n", floor,chip);
    return 0;
  }
  
  if(DoC_Command(doc->virtadr, 0x90)) {
    printk("DoC_Command (ReadID) for %d,%d returned true\n", floor,chip);
    return 0;
  }
  DoC_Address(doc->virtadr, 0);
  
  mfr = ReadDOC(doc->virtadr, 2k_CSDN_IO);
  id = ReadDOC(doc->virtadr, 2k_CSDN_IO);

  /* No response - return failure */
  if (mfr == 0xff || mfr == 0)
    return 0;

  if (doc->mfr) {
    if (doc->mfr == mfr && doc->id == id)
      return 1; /* This is another the same the first */
    else
      printk(KERN_WARNING "Flash chip at floor %d, chip %d is different:\n",
	     floor, chip);
  }
  
  /* This is the first time. Print and store the manufacturer and ID codes. */
  
  switch(mfr) {
  case 0x98: /* Toshiba */
    mfrname = "Toshiba";
    
    switch(id)
      { /* How many of these actually get used in DiskOnChip */
      case 0xea:
	idname = "TC58V16BDC";
	chipshift = 21;
	break;

      case 0x64:
	idname = "TC5816BDC";
	chipshift = 21;
	break;

      case 0x6b:
	idname = "TC5832DC";
	chipshift = 22;
	break;

      case 0xe5:
	idname = "TC58V32DC";
	chipshift = 22;
	break;

      case 0xe6: 
	idname = "TC58V64DC";
	chipshift = 23;
	break;

      case 0x73: 
	idname = "TH58V128DC";
	chipshift = 24;
	break;

      case 0x75: 
	idname = "TC58256FT/DC";
	chipshift = 25;
	break;
      }
    break; /* End of Toshiba parts */

  case 0xec: /* Samsung */
    mfrname = "Samsung";

    switch(id) {
    case 0x64:
      idname = "KM29N16000";
      chipshift = 21;

    case 0xea:
      idname = "KM29W16000";
      chipshift = 21;
      break;

    case 0xe3:
      idname = "KM29W32000";
      chipshift = 22;
      break;

    case 0xe6:
      idname = "KM29U64000";
      chipshift = 23;
      break;
      
    case 0x73:
      idname = "KM29U128T";
      chipshift = 24;
      break;

    case 0x75:
      idname = "KM29U256T";
      chipshift = 25;
      break;
    }
    break; /* End of Samsung parts */
  }
  
  if (idname) {
    printk (KERN_INFO "Flash chip found: %2.2X %2.2X (%s %s)\n", 
	    mfr,id,mfrname,idname);
    if (!doc->mfr) {
      doc->mfr = mfr;
      doc->id = id;
      doc->chipshift = chipshift;
      return 1;
    }
    return 0;
  }

  /* OK. We haven't fully identified the chip. */

  if (mfrname)
    printk(KERN_INFO "Unknown %s flash chip found: %2.2X %2.2X\n", mfrname,
	   id, mfr);
  else
    printk(KERN_WARNING "Unknown flash chip found: %2.2X %2.2X\n", id, mfr);

  printk(KERN_WARNING "Please report to David.Woodhouse@mvhi.com\n");
  return 0;
}     

void DoC_ScanChips(struct DiskOnChip *this)
{
  int floor, chip;
  int numchips[MAX_FLOORS];
  int ret = 1;

  this->numchips = 0;
  this->mfr = 0;
  this->id = 0;
  for (floor = 0 ; floor < MAX_FLOORS ; floor++) {
    ret = 1;
    numchips[floor]=0;
    for (chip = 0 ; chip < MAX_CHIPS && ret != 0; chip++ ) {
      
      ret = DoC_IdentChip(this, floor, chip);
      if (ret) {
	numchips[floor]++;
	this->numchips++;
      }
    }
  }

  if (!this->numchips) {
    printk("No flash chips recognised.\n");
    return;
  }

  this->chips = kmalloc(sizeof(struct Nand) * this->numchips, GFP_KERNEL);
  if (!this->chips){
    printk("No memory for allocating chip info structures\n");
    return;
  }

  ret = 0;

  for (floor = 0; floor < MAX_FLOORS; floor++) {
    for (chip = 0 ; chip < numchips[floor] ; chip++) {
      this->chips[ret].floor = floor;
      this->chips[ret].chip = chip;
      this->chips[ret].curadr = 0;
      this->chips[ret].curmode = 0x50;
      ret++;
    }
  }

  this->totlen = this->numchips * (1 << this->chipshift);
    printk("%d chips found. Total DiskOnChip size: %ld Mb\n", this->numchips ,
	   this->totlen >> 20);
    
}


void DoC_Probe(unsigned long physadr)
{
  unsigned long docptr = (unsigned long)ioremap(physadr, 0x2000);
  struct DiskOnChip *this;
  struct mtd_info *mtd;
  int ChipID;

  if (!docptr) return;
  
  if ((ChipID = doccheck(docptr, physadr)))
    {
      mtd = kmalloc(sizeof(struct DiskOnChip) + sizeof(struct mtd_info), GFP_KERNEL);

      if (!mtd) {
	printk("Cannot allocate memory for data structures. Dropping.\n");
	return;
      }
	
      this = (struct DiskOnChip *)(&mtd[1]);
      
      memset((char *)mtd,0, sizeof(struct mtd_info));

      mtd->priv = this;
      strncpy(mtd->name, "DiskOnChip 2000", 31);
      mtd->type = MTD_NANDFLASH;
      mtd->flags = MTD_CAP_NANDFLASH;
      mtd->size = 0;
      mtd->erasesize = 0x2000;
      mtd->oobblock = 512;
      mtd->oobsize = 16;
#ifdef MODULE
      mtd->module = &__this_module;
#endif
      mtd->point = NULL;
      mtd->unpoint = NULL;
      mtd->read = doc_read;
      mtd->write = doc_write;
      mtd->read_oob = doc_read_oob;
      mtd->write_oob = NULL; //doc_writeoob;
      mtd->sync = NULL;
      
      this->physadr = physadr;
      this->virtadr = docptr;
      this->ChipID = ChipID;

      this->totlen = 0;
      this->numchips = 0;

      this->curfloor = -1;
      this->curchip = -1;
      /* Ident all the chips present. */
      DoC_ScanChips(this);
      
      if (!this->totlen)
	kfree(mtd);
      else {
	numdocs++;
	mtd->size  = this->totlen;
	add_mtd_device(mtd);
      }
    }
}
      


int doccheck(unsigned long potential, unsigned long physadr)
{
  unsigned long window=potential;
  unsigned char tmp, tmp2, ChipID;
  

  /* Routine copied from the Linux DOC driver */

  if (readb(window) != 0x55 || readb(window+1) != 0xaa)
    {
      //printk ("Start of window not 0x55 0xaa: %2.2X %2.2X\n", 
      //      (unsigned int)readb(window),(unsigned int)readb(window+1));
      return 0;
    }

  tmp2 = ReadDOC(window, DOCControl);
  
  WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_RESET, 
	   window, DOCControl);
  WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_RESET, 
	   window, DOCControl);
  WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_NORMAL, 
	   window, DOCControl);
  WriteDOC(DOC_MODE_CLR_ERR | DOC_MODE_MDWREN | DOC_MODE_NORMAL, 
	   window, DOCControl);
  
  ChipID = ReadDOC(window, ChipID);
  
  switch (ChipID) {
  case DOC_ChipID_Doc2k:
    printk(KERN_INFO "Disk-On-Chip 2000 found at 0x%lx\n",physadr);

    tmp = ReadDOC(window, 2k_Toggle) & 4;
    if ((ReadDOC(window, 2k_Toggle) & 4) != tmp)
      {
	return ChipID;
      }
    break;

  case DOC_ChipID_DocMil:
    printk(KERN_INFO "Disk-On-Chip Millennium found at 0x%lx\n",physadr);
    /* We don't actually handle the MDOC yet - it's all hard-coded for the DOC.
       That'll change when I actually have a MDOC to play with */
    tmp = ReadDOC(window, ECCConf) & 4;
    if ((ReadDOC(window, ECCConf) & 4) != tmp)
      {
	return ChipID;
      }
    break;

  default:
    WriteDOC(tmp2, window, DOCControl);
    printk(KERN_WARNING "Disk-On-Chip with unknown ChipID %2.2X found at 0x%lx\n",
	   ChipID, physadr);
    return 0;
  }

  printk(KERN_WARNING "DiskOnChip failed TOGGLE test, dropping.\n");
  WriteDOC(tmp2, window, DOCControl);
  return 0;
}   

static int doc_read (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf)
{
  struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
  size_t di=0;
  int num = 0; //MINOR(filp->f_dentry->d_inode->i_rdev); Must start using
               //the right device. 
  unsigned long docptr;
  struct Nand *mychip;
  loff_t ofs;

  if (num >= numdocs)
    return -ENODEV;

  docptr = this->virtadr;

  if (from > this->totlen)
    return -EINVAL;

  if (from + len >= this->totlen)
    len = this->totlen - from;

  // Enforce only a single chip. Urgh. How to simply make a 'mask of %d bits'?
  if ((from >> this->chipshift) != (from + len) >> this->chipshift)
    len = (((from >> this->chipshift) + 1 ) << this->chipshift ) - from;

  mychip = &this->chips[from >> (this->chipshift)];

  if (this->curfloor != mychip->floor) {
    DoC_SelectFloor(docptr, mychip->floor);
    DoC_SelectChip(docptr, mychip->chip);
  }
  else if (this->curchip != mychip->chip) {
    DoC_SelectChip(docptr, mychip->chip);
  }

  this->curfloor = mychip->floor;
  this->curchip = mychip->chip;

  /* Yes we don't guarantee to set the address. I'll fix it later */

  for (ofs = from; ofs < from + len ; ofs++)
    {
      if (!(ofs & 0xff)) {
	/* Some optimisation required - cache the command and address so we
	   don't have to load them _every_ time. Also select the correct chip,
	   for multi-chip DOCs. The latter will happen when I have one. */
	DoC_Command(docptr, (ofs >> 8) & 1);
	//		DoC_Command(docptr, 0x50);
	DoC_Address(docptr, ofs);
      }

      //(void)readb(docptr + 0x100d);
      buf[di++] = ReadDOC(docptr, 2k_CSDN_IO);
    }

  *retlen = len;
  return 0;
}

static ssize_t doc_write (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf)
{
  /* Yeah, right */

  return -EIO;
}


static int doc_release(struct inode *inode, struct file *file)
{
  return 0;
}

static int doc_open(struct inode *inode, struct file *filp)
{
  return 0;
}


static int doc_read_oob(struct mtd_info *mtd, loff_t ofs, size_t len, size_t *retlen, u_char *buf)
{
  struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
  int i;
  unsigned long docptr;
  struct Nand *mychip;


  docptr = this->virtadr;

  mychip = &this->chips[ofs >> this->chipshift];

  if (this->curfloor != mychip->floor) {
    DoC_SelectFloor(docptr, mychip->floor);
    DoC_SelectChip(docptr, mychip->chip);
  }
  else if (this->curchip != mychip->chip) {
    DoC_SelectChip(docptr, mychip->chip);
  }
  this->curfloor = mychip->floor;
  this->curchip = mychip->chip;



  DoC_Command(docptr, 0x50);
  DoC_Address(docptr, ofs);
  
  for (i=0; i<len; i++)
      buf[i] = ReadDOC(docptr, 2k_CSDN_IO);

  *retlen = len;
  return 0;

}



/****************************************************************************
 *
 * Module stuff
 *
 ****************************************************************************/



int init_module(void)
{
  int i;
  

  printk("M-Systems DiskOnChip driver. (C) 1999 Machine Vision Holdings, Inc.\n");
  printk("$Id: doc2000.c,v 1.1 1999/08/09 19:38:05 dwmw2 Exp $\n");

  for (i=0; doc_locations[i]; i++) {
    DoC_Probe(doc_locations[i]);
  }

  if (numdocs)
    return 0;

  printk("No Disk-On-Chip hardware found\n");
  return -ENXIO;
    
}


void cleanup_module(void)
{
  int i;
  struct mtd_info mtd;

  for (i=0; i < MAX_MTD_DEVICES; i++) {
    struct mtd_info *mtd = get_mtd_device(i);

    if (mtd && mtd->module == &__this_module) {
      struct DiskOnChip *this = (struct DiskOnChip *)mtd->priv;
      
      del_mtd_device(mtd);
      
      iounmap((void *)this->virtadr);
      kfree(this->chips);
      kfree(mtd);
    }
  }
}















