
/* Linux driver for Disk-On-Chip 2000       */
/* (c) 1999 Machine Vision Holdings, Inc.   */
/* Author: David Woodhouse <dwmw2@mvhi.com> */
/* $Id: nftl.c,v 1.4 1999/08/10 15:52:24 dwmw2 Exp $ */

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/sched.h>

#include "mtd.h"
#include "nftl.h"

/* NFTL block device stuff */
#define MAJOR_NR NFTL_MAJOR
#define DEVICE_REQUEST nftl_request
#define DEVICE_OFF(device)
#include <linux/blk.h>
#include <linux/hdreg.h>


#define MAX_NFTLS 16

/* Given a pointer to a NFTL Media Header, initialise all the structures and
   register the block device */
struct NFTLrecord {
  struct mtd_info *mtd;
  u16 MediaUnit, SpareMediaUnit;
  u32 EraseSize;
  struct NFTLMediaHeader MediaHdr;
  u16 numvunits;
  u32 long nr_sects;
  int head,sect,cyl;
#define numEUNs MediaHdr.NumEraseUnits
  u16 *EUNtable; /* [numvunits]: First EUN for each virtual unit  */
  u16 *VirtualUnitTable; /* [numEUNs]: VirtualUnitNumber for each */
  u16 *ReplUnitTable; /* [numEUNs]: ReplUnitNumber for each */
} *NFTLs[MAX_NFTLS];

int numnftls = 0;

void NFTL_setup(struct mtd_info *mtd, unsigned long ofs, 
			 struct NFTLMediaHeader *hdr)
{
  int i;
  struct NFTLrecord *thisNFTL;

  for (i=0; i < numnftls; i++)
    if (NFTLs[i] && NFTLs[i]->mtd == mtd && 
	NFTLs[i]->MediaHdr.FirstPhysicalEUN == hdr->FirstPhysicalEUN) {
      /* This is a Spare Media Header for an NFTL we've already found */
      //      printk("Spare Media Header for NFTL %d found at %lx\n",i, ofs);
      NFTLs[i]->SpareMediaUnit = ofs / mtd->erasesize;
      return;
    }

  

  /* OK, it's a new one. Set up all the data structures. */

  printk("Found new NFTL nftl%c at offset %lx\n",numnftls + 'a', ofs);

  if (hdr->UnitSizeFactor != 0xff) {
    printk("Sorry, we don't support UnitSizeFactor of != 1 yet\n");
    return;
  }

  thisNFTL = kmalloc(sizeof(struct NFTLrecord), GFP_KERNEL);
  if (!thisNFTL) {
    printk(KERN_WARNING "Out of memory for NFTL data structures\n");
    return;
  }

  thisNFTL->EraseSize = mtd->erasesize;
  memcpy(&thisNFTL->MediaHdr, hdr, sizeof(*hdr));
  thisNFTL->mtd = mtd;
  thisNFTL->MediaUnit = ofs / mtd->erasesize;
  thisNFTL->SpareMediaUnit = 0xffff;
  thisNFTL->numvunits = le32_to_cpu(thisNFTL->MediaHdr.FormattedSize) / 8192;
  thisNFTL->nr_sects = thisNFTL->numvunits  * (thisNFTL->EraseSize / 512);


  thisNFTL->EUNtable = kmalloc( 2 * thisNFTL->numvunits,
			       GFP_KERNEL);
  if (!thisNFTL->EUNtable) {
    printk("ENOMEM\n");
    kfree(thisNFTL);
    return;
  }
  memset(thisNFTL->EUNtable, 0xff, 2 * thisNFTL->numvunits);

  thisNFTL->VirtualUnitTable = kmalloc( 2 * thisNFTL->numEUNs , GFP_KERNEL);
  if (!thisNFTL->VirtualUnitTable) {
    printk("ENOMEM\n");
    kfree(thisNFTL->EUNtable);
    kfree(thisNFTL);
    return;
  }
  memset(thisNFTL->VirtualUnitTable, 0xff, 2 * thisNFTL->numEUNs);
    
  thisNFTL->ReplUnitTable = kmalloc( 2 * thisNFTL->numEUNs , GFP_KERNEL);
  if (!thisNFTL->ReplUnitTable) {
    printk("ENOMEM\n");
    kfree(thisNFTL->VirtualUnitTable);
    kfree(thisNFTL->EUNtable);
    kfree(thisNFTL);
    return;
  }
  memset(thisNFTL->ReplUnitTable, 0xff, 2 * thisNFTL->numEUNs);
  
  /* Ought to check the media header for bad blocks */


  /* Scan each physical Erase Unit for validity and to find the 
     Virtual Erase Unit Chain to which it belongs */

  for (i=le16_to_cpu(thisNFTL->MediaHdr.FirstPhysicalEUN); 
       i < thisNFTL->numEUNs + 
	 le16_to_cpu(thisNFTL->MediaHdr.FirstPhysicalEUN); i++) {

    union doc_uci uci;
    unsigned long ofs;
    size_t retlen;
    ofs = i * thisNFTL->EraseSize;

    MTD_READOOB(mtd, (i * thisNFTL->EraseSize) + 512 + 8, 8, &retlen, (char *)&uci);

    if (uci.b.EraseMark != cpu_to_le16(0x3c69) || 
	uci.b.EraseMark1 != cpu_to_le16(0x3c69)) {
      printk("EUN %d: EraseMark not 0x3c69 (0x%4.4x 0x%4.4x instead)\n",
	     i, le16_to_cpu(uci.b.EraseMark), le16_to_cpu(uci.b.EraseMark1));
      continue;
    }
    
    MTD_READOOB(mtd, (i * thisNFTL->EraseSize) + 8, 8, &retlen, (u_char *)&uci);
    
    if (uci.a.VirtUnitNum != uci.a.SpareVirtUnitNum)
      printk("EUN %d: VirtualUnitNumber (%x) != SpareVirtualUnitNumber (%x)\n",
	     i, le16_to_cpu(uci.a.VirtUnitNum), 
	     le16_to_cpu(uci.a.SpareVirtUnitNum));

    if (uci.a.ReplUnitNum != uci.a.SpareReplUnitNum)
      printk("EUN %d: ReplacementUnitNumber (%x) != SpareReplacementUnitNumber (%x)\n",
	     i, le16_to_cpu(uci.a.ReplUnitNum), 
	     le16_to_cpu(uci.a.SpareReplUnitNum));

    /* We don't actually _do_ anything about the above, just whinge */

    thisNFTL->VirtualUnitTable[i] = le16_to_cpu(uci.a.VirtUnitNum);
    thisNFTL->ReplUnitTable[i] = le16_to_cpu(uci.a.ReplUnitNum);


    /* if (!(VUN & 0x8000) && VUN < (arraybounds)).. optimises to: */
    if (le16_to_cpu(uci.a.VirtUnitNum) < thisNFTL->numvunits) 
      thisNFTL->EUNtable[le16_to_cpu(uci.a.VirtUnitNum) & 0x7fff] = i;

  } 
  NFTLs[numnftls++] = thisNFTL;

  if (thisNFTL->mtd->module)
    __MOD_INC_USE_COUNT(thisNFTL->mtd->module);
  
#if 0
  for (i=0; i < 10/* thisNFTL->numvunits*/; i++) {
    u16 curEUN = thisNFTL->EUNtable[i];
    int sillycount=100;

    printk("Virtual Unit #%d: ",i);
    if (!curEUN || curEUN == 0xffff) {
      printk("Not present\n");
      continue;
    }
    printk("%d", curEUN);

    while ((curEUN = thisNFTL->ReplUnitTable[curEUN]) != 0xffff && --sillycount) {
      printk(", %d", curEUN & 0xffff);
      
    }
      printk("\n");
  }
#endif
}


/* Search each DiskOnChip for NFTL partitions */

void NFTL_init(void)
{
  int i;
  struct mtd_info *mtd;
  unsigned long ofs;
  struct NFTLMediaHeader hdr;

  for (i=0; i < MAX_MTD_DEVICES; i++) {
    mtd = get_mtd_device(i);

    if (!mtd) 
      continue;

    if (!mtd->read_oob) /* If this MTD doesn't have out-of-band data,
			   then there's no point continuing */
      continue;

    for (ofs = 0; ofs < mtd->size ; ofs += mtd->erasesize) {
	size_t retlen;
	MTD_READ(mtd, ofs, sizeof(hdr), &retlen, (u_char *)&hdr);
	
	if (retlen < sizeof(hdr))
	  continue;
	
	if (!strncmp(hdr.DataOrgID, "ANAND", 6)) {
	  
	  NFTL_setup(mtd, ofs, &hdr);
	}
      }
  }
}


void dummy_init (struct gendisk *crap)
{}


static int nftl_ioctl(struct inode * inode, struct file * file,
                    unsigned int cmd, unsigned long arg)
{
  printk("NFTL ioctl called\n");
  return -EINVAL;
}

struct hd_struct part_table[256] = {{0,0},};
static int nftl_sizes[256]={0,};
static int nftl_blocksizes[256] = {0,};

void nftl_request(void)
{
  unsigned int dev, block, nsect, blockofs;
  struct NFTLrecord *thisNFTL;
  char *buffer;
repeat:
  cli();
  INIT_REQUEST;
  dev = MINOR(CURRENT->rq_dev);
  block = CURRENT->sector;
  nsect = CURRENT->nr_sectors;
  buffer = CURRENT->buffer;

  if (dev >= numnftls * 16) {
    printk("fl: bad minor number: device=%s\n",
	   kdevname(CURRENT->rq_dev));
    end_request(0); /* fail */
    goto repeat;
  }
  
  thisNFTL = NFTLs[dev / 16];
  
  if (block + nsect >= part_table[dev].nr_sects) {
    printk("nftl%c%d: bad access: block=%d, count=%d\n",
	   (MINOR(CURRENT->rq_dev)>>6)+'a', dev & 0xf, block, nsect);
    end_request(0); /* fail */
    goto repeat;
  }

  block += part_table[dev].start_sect;
  
  if (CURRENT->cmd == READ) {


    for ( ; nsect > 0; nsect-- , block++, buffer+= 512) {
      /* Read a single sector to CURRENT->buffer + (512 * i) */
      u16 lastgoodEUN = 0xffff;
      u16 thisEUN = thisNFTL->EUNtable[block / (thisNFTL->EraseSize / 512)];
      blockofs = (block * 512) & (thisNFTL->EraseSize -1);

      if (thisEUN == 0xffff) thisEUN = 0;
      
      while(thisEUN && (thisEUN & 0x7fff) != 0x7fff) {
	struct doc_bci bci;
	size_t retlen;

	MTD_READOOB(thisNFTL->mtd, (thisEUN * thisNFTL->EraseSize) + blockofs,8, &retlen, (char *)&bci);

	switch(bci.Status) {
	case cpu_to_le16(BLOCK_FREE):
	  thisEUN = 0;
	  break;
	case cpu_to_le16(BLOCK_USED):
	  lastgoodEUN = thisEUN;
	  break;
	default:
	  printk("Unknown status for block %d in EUN %d: %x\n",block,thisEUN, bci.Status);
	}
	if (thisEUN)
	  thisEUN = thisNFTL->ReplUnitTable[thisEUN] & 0x7fff;
      }
      if (lastgoodEUN == 0xffff) {
	memset(buffer, 0, 512);
      } else {
	loff_t ptr = (lastgoodEUN * thisNFTL->EraseSize) + blockofs;
	size_t retlen;
	MTD_READ(thisNFTL->mtd, ptr, 512, &retlen, buffer);
      }
    }
	
    end_request(1);
    goto repeat;
    }
  else if (CURRENT->cmd == WRITE) {
    printk("Ouch, gotta write request\n");
    end_request(0); /* fail */
    goto repeat;
  }
  else {

    end_request(0); /* fail */
    goto repeat;
  }
}

static int nftl_open(struct inode *ip, struct file *fp)
{
  if (fp->f_mode & FMODE_WRITE)
    return -EROFS;

  MOD_INC_USE_COUNT;
  return 0;
}

static int nftl_release(struct inode *ip, struct file *fp)
{
  MOD_DEC_USE_COUNT;
  return 0;
}

static struct gendisk nftl_gendisk = {
        NFTL_MAJOR,     /* Major number */      
        "nftl",         /* Major name */
        4,              /* Bits to shift to get real from partition */
        15,             /* Number of partitions per real */
        MAX_NFTLS,      /* maximum number of real */
        dummy_init,     /* init function */
        part_table,     /* hd struct */
        nftl_sizes,     /* block sizes */
        0,              /* number */
        NULL,           /* internal use, not presently used */
        NULL            /* next */
};

static struct file_operations nftl_fops = {
        NULL,           /* lseek - default */
        block_read,     /* read - block dev read */
        block_write,    /* write - block dev write */
        NULL,           /* readdir - not here! */
        NULL,           /* select */
        nftl_ioctl,     /* ioctl */
        NULL,           /* mmap */
        nftl_open,           /* open */
        NULL,           /* flush */
        nftl_release,           /* no special release code... */
        block_fsync     /* fsync */
};




/****************************************************************************
 *
 * Module stuff
 *
 ****************************************************************************/



int init_module(void)
{
  int i;
  

  printk("M-Systems NAND Flash Translation Layer driver.\n (C) 1999 Machine Vision Holdings, Inc.\n");
  printk("$Id: nftl.c,v 1.4 1999/08/10 15:52:24 dwmw2 Exp $\n");

  NFTL_init();
  if (numnftls) {
    
    for (i=0; i < numnftls ; i++) {
      part_table[i * 16].nr_sects = NFTLs[i]->nr_sects;
    }
    
    if (register_blkdev(NFTL_MAJOR, "nftl", &nftl_fops)){
    printk("unable to register NFTL block device\n");
    } else {
      blk_dev[NFTL_MAJOR].request_fn = DEVICE_REQUEST;
      for (i=0; i < 256 ; i++)
	nftl_blocksizes[i] = 1024;
      blksize_size[NFTL_MAJOR] = nftl_blocksizes;
      nftl_gendisk.next = gendisk_head;
     gendisk_head = &nftl_gendisk;
    }

    for (i=0; i < numnftls ; i++) {
      resetup_one_dev(&nftl_gendisk, i);
    }

    return 0;
  }
  printk("No NFTL partitions found\n");
  return -ENODEV;
}




void cleanup_module(void)
{
  struct gendisk *gd, **gdp;


  for (numnftls-- ; numnftls >= 0; numnftls--) {
    kfree(NFTLs[numnftls]->VirtualUnitTable);
    kfree(NFTLs[numnftls]->ReplUnitTable);
    kfree(NFTLs[numnftls]->EUNtable);
    if (NFTLs[numnftls]->mtd->module)
    __MOD_DEC_USE_COUNT(NFTLs[numnftls]->mtd->module);
    kfree(NFTLs[numnftls]);
  }

  unregister_blkdev(NFTL_MAJOR, "nftl");
  for (gdp = &gendisk_head; *gdp; gdp = &((*gdp)->next))
    if (*gdp == &nftl_gendisk) {
      gd = *gdp; *gdp = gd->next;
      break;
    }
  
}
















