// -*- mode: cpp; mode: fold -*-
// Description                                                          /*{{{*/
// $Id: octagon-5066.c,v 1.2 1999/08/09 18:32:14 dwmw2 Exp $
/* ######################################################################

   Octagon 5066 MTD Driver. 
  
   The Octagon 5066 is a SBC based on AMD's 586-WB running at 133 MHZ. It
   comes with a builtin AMD 29F016 flash chip and a socketed EEPROM that
   is replacable by flash. Both units are mapped through a multiplexer
   into a 32k memory window at 0xe8000. The control register for the 
   multiplexing unit is located at IO 0x208 with a bit map of
     0-5 Page Selection in 32k increments
     6-7 Device selection:
        00 SSD off
        01 SSD 0 (Socket)
        10 SSD 1 (Flash chip)
        11 undefined
  
   On each SSD, the first 128k is reserved for use by the bios
   (actually it IS the bios..) This only matters if you are booting off the 
   flash, you must not put a file system starting there.
   
   The driver tries to do a detection algorithm to guess what sort of devices
   are plugged into the sockets.
   
   ##################################################################### */
									/*}}} */
// Include Files                                                        /*{{{*/
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <asm/io.h>

#include "mapped.h"
									/*}}} */

#define WINDOW_START 0xe8000
#define WINDOW_LENGTH 0x8000
#define PAGE_IO 0x208

static struct mapped_mtd_info *SSD[2];
static u_char WindowPos = 0;
static int OctLock = 0;

// oct5066_page - Return a readb address for the new page window	/*{{{*/
// ---------------------------------------------------------------------
/* All page windows are of the page window size set at map creation, 
   map->pagesize. All this function does is move the page window and 
   return the address */
static unsigned long oct5066_page(struct mapped_mtd_info *map,
				  unsigned long page)
{
   unsigned char NewPos;
   unsigned long Device = (unsigned long)map->mtd.priv;
   if (page > 0x3F)
      return 0;
      
   // See if we need to move the window
   NewPos = Device | page;
   if (WindowPos != NewPos)
   {
      outb(NewPos,PAGE_IO);
      WindowPos = NewPos;
   }
   return WINDOW_START;
}
									/*}}}*/

// CreateSSD - Construct the SSD structure				/*{{{*/
// ---------------------------------------------------------------------
/* Since two SSDs are created this simplifies the init code.. */
static struct mapped_mtd_info *CreateSSD()
{
   // Allocate some memory
   struct mapped_mtd_info *mtd;
   mtd = (struct mapped_mtd_info *)kmalloc(sizeof(*mtd),GFP_KERNEL);
   if (mtd == 0)
      return 0;   
   memset(mtd,0,sizeof(*mtd));

   // Setup the MTD structure
#ifdef MODULE
   mtd->mtd.module = &__this_module;
#endif
   mtd->page = oct5066_page;
   mtd->pagesize = WINDOW_LENGTH;
   mtd->maxsize = 0x3F;
   
   /* Give it a lock for our driver instance. This will prevent ANY 
      concurrent activity between the two flash devices using the window */
   mtd->mtd.lock = &OctLock;
   
   return mtd;
}
									/*}}}*/
// OctProbe - Sense if this is an octagon card				/*{{{*/
// ---------------------------------------------------------------------
/* Perform a simple validity test, we map the window select SSD0 and
   change pages while monitoring the window. A change in the window, 
   controlled by the PAGE_IO port is a functioning 5066 board. This will
   fail if the thing in the socket is set to a uniform value. */
static int __init OctProbe()
{
   unsigned int Base = (1 << 6);
   unsigned long I;
   unsigned long Values[10];
   for (I = 0; I != 20; I++)
   {
      outb(Base + (I%10),PAGE_IO);
      if (I < 10)
      {
	 // Record the value and check for uniqeness
	 Values[I%10] = readl(WINDOW_START);
	 if (I > 0 && Values[I%10] == Values[0])
	    return -EAGAIN;
      }      
      else
      {
	 // Make sure we get the same values on the second pass
	 if (Values[I%10] != readl(WINDOW_START))
	    return -EAGAIN;
      }      
   }
   return 0;
}
									/*}}}*/

// cleanup_module - Clean up the allocations made			/*{{{*/
// ---------------------------------------------------------------------
/* */
#ifdef MODULE
#define octagon5066_cleanup cleanup_module
#endif
void octagon5066_cleanup()
{
   int I;
   for (I = 0; I != 2; I++)
   {
      mtd_mapped_remove(SSD[I]);
      kfree(SSD[I]);
      SSD[I] = 0;
   }
   release_region(PAGE_IO,1);
}
									/*}}}*/
// init_module - Module startup function				/*{{{*/
// ---------------------------------------------------------------------
/* Create the two SSD devices and verify that this is an Octagon SBC */
#ifdef MODULE
#define octagon5066_init init_module
#endif   
int __init octagon5066_init()
{
   // Do an autoprobe sequence
   if (check_region(PAGE_IO,1) != 0)
   {
      printk("5066: Page Register in Use\n");
      return -EAGAIN;
   }

   if (OctProbe() != 0)
   {
      printk("5066: Octagon Probe Failed, is this an Octagon 5066 SBC?\n");
      return -EAGAIN;
   }

   request_region(PAGE_IO,1,"Octagon SSD");

   // Print out our little header..
   printk("Octagon 5066 SSD IO:0x%x MEM:0x%x-0x%x\n",PAGE_IO,WINDOW_START,
	  WINDOW_START+WINDOW_LENGTH);
   	  
   // Create the SSDs
   SSD[0] = CreateSSD();
   SSD[1] = CreateSSD();
   if (SSD[0] == 0 || SSD[1] == 0)
   {
      octagon5066_cleanup();
      return -EAGAIN;
   }

   /* We set characteristics for two ROM devices of 512 and 2M, I don't know
      how to sense ROM size yet.. */
   strcpy(SSD[0]->mtd.name,"5066 Socket");
   SSD[0]->mtd.size = 512*1024;
   SSD[0]->mtd.type = MTD_ROM;
   (unsigned long)SSD[0]->mtd.priv = 1 << 6;
   strcpy(SSD[1]->mtd.name,"5066 Internal");
   SSD[1]->mtd.size = 2*1024*1024;
   SSD[1]->mtd.type = MTD_ROM;
   (unsigned long)SSD[1]->mtd.priv = 2 << 6;

   // If either are flash then the correct settings will be updated now..
   if (mtd_mapped_setup(SSD[0]) != 0 || mtd_mapped_setup(SSD[1]) != 0)
   {
      printk("Failed to register new device\n");
      octagon5066_cleanup();
      return -EAGAIN;
   }
   
   return 0;
}
									/*}}}*/
